

#include <iostream>
#include <stdexcept>
#include <sstream>
using namespace std;

#undef NDEBUG

#include "config.h"

extern "C"{
#include <ascend/general/platform.h>

#include <ascend/general/list.h>
#include <ascend/compiler/ascCompiler.h>

/* #include <compiler/redirectFile.h> */
#include <ascend/compiler/prototype.h>
#include <ascend/compiler/dump.h>
#include <ascend/compiler/childio.h>
#include <ascend/compiler/type_desc.h>
#include <ascend/compiler/typedef.h>
#include <ascend/compiler/library.h>
#include <ascend/system/slv_types.h>
#include <ascend/system/system.h>
#include <ascend/utilities/ascEnvVar.h>
#include <ascend/compiler/symtab.h>
#include <ascend/general/table.h>
#include <ascend/compiler/instance_enum.h>
#include <ascend/compiler/notate.h>
#include <ascend/compiler/simlist.h>
#include <ascend/compiler/parser.h>
#include <ascend/utilities/error.h>
#include <ascend/general/env.h>
#include <ascend/compiler/importhandler.h>
}

#include "library.h"
#include "simulation.h"
#include "solver.h"

//#define ASCXX_LIBRARY_DEBUG
#ifdef ASCXX_LIBRARY_DEBUG
# define MSG CONSOLE_DEBUG
#else
# define MSG(ARGS...) ((void)0)
#endif

Library::Library(const char *defaultpath){
	static int have_init;
	if(!have_init){
		MSG("Initialising ASCEND library...");

#ifdef REIMPLEMENT_STREAMS
		Asc_RedirectCompilerDefault(); // Ensure that error message reach stderr
#endif

		Asc_CompilerInit(1);
		env_import(ASC_ENV_LIBRARY,getenv,Asc_PutEnv);
		env_import(ASC_ENV_SOLVERS,getenv,Asc_PutEnv);

		char *x = Asc_GetEnv(ASC_ENV_LIBRARY);
		if(x==NULL || strcmp(x,"")==0){
			if(defaultpath==NULL){
				ERROR_REPORTER_NOLINE(ASC_PROG_WARNING,"Using default "
					ASC_ENV_LIBRARY " = '" DEFAULT_ASCENDLIBRARY "'"
				);
				defaultpath = DEFAULT_ASCENDLIBRARY;
			}

			string s = string(ASC_ENV_LIBRARY "=") + defaultpath;
			//ERROR_REPORTER_HERE(ASC_PROG_NOTE,"Setting %s",s.c_str());;
			Asc_PutEnv(s.c_str());
		}
		Asc_ImportPathList(ASC_ENV_LIBRARY);
		//cerr << PATHENVIRONMENTVAR << " = " << x << endl;
		//cerr << "Created LIBRARY" << endl;
		//cerr << "Registering solvers..." << endl;
		registerStandardSolvers();
	}
	have_init=1;
}

Library::~Library(){
	//ERROR_REPORTER_NOLINE(ASC_PROG_WARNING,"DESTROYED LIBRARY!////////////////////////////////////////");
	//DestroyLibrary();
	// ... need to use some kind of reference counting before you can do that...
}

/**
	Load an ASCEND model file into the Library. It will be parsed such that
	its types will be visible to Library::findType.

	@param filename Filename, will be searched for relative to ASCENDLIBRARY environment
		variable, if necessary.
*/
void
Library::load(const char *filename){

	//std::cerr << "Loading '" << filename << "'" << std::endl;

	int status;
	struct module_t *m=Asc_RequireModule(filename,&status);
	if(m!=NULL){
		//std::cerr << "Loaded module '" << Asc_ModuleName(m) << "'" << std::endl;
	}else{
		std::cerr << "Error: unable to load module '" << filename << "'." << std::endl;
	}

	const char *msg = getLoadErrorMessage(status);

	char msg1[100];
	sprintf(msg1,msg,filename);

	if(status<0 || status>0){
		throw std::runtime_error(msg1);
#ifdef ASCXX_LIBRARY_DEBUG
	}else{
		MSG(msg1,Asc_ModuleName(m));
#endif
	}

	MSG("Beginning parse of %s",Asc_ModuleName(m));
	error_reporter_tree_t *tree1 = error_reporter_tree_start(0);
	status = zz_parse();
	switch(status){
		case 0: break;
		case 1: ERROR_REPORTER_NOLINE(ASC_USER_ERROR,"Parsing of %s was aborted",Asc_ModuleName(m)); break;
		case 2: ERROR_REPORTER_NOLINE(ASC_PROG_FATAL,"Out of memory when parsing %s",Asc_ModuleName(m)); break;
		default: ERROR_REPORTER_NOLINE(ASC_PROG_ERROR,"Invalid return from zz_parse"); break;
	}
	status = error_reporter_tree_has_error(tree1);
	if(!status){
		error_reporter_tree_end_clear(tree1);
	}else{
		error_reporter_tree_end(tree1);
		ERROR_REPORTER_NOLINE(ASC_USER_ERROR,"Error(s) when loading '%s'",filename);
		stringstream ss;
		ss << "Errors found in '" << filename <<  "'";
		throw runtime_error(ss.str());
	}

#ifdef ASCXX_LIBRARY_DEBUG
	struct gl_list_t *l = Asc_TypeByModule(m);
	MSG("%lu library entries loaded from %s",gl_length(l), filename);
#endif
}

/**
	Load MODELs specified inside a text string into the Library. The string is
	parsed and its types will then be visible to Library::findType.

	@param str String containing the model, in the ASCEND modelling language.
	@param nameprefix Name to be used to refer tothe string module when eg using
		the Library::getModules methood.
*/
void
Library::loadString(const char *str, const char *nameprefix){
	int status;
	struct module_t *m = Asc_OpenStringModule(str, &status, nameprefix);

	const char *msg = getLoadErrorMessage(status);

	char msg1[100];
	sprintf(msg1,msg,nameprefix);

	if(status<0 || status>0){
		throw std::runtime_error(msg1);
	}else{
		std::cerr << "Note: Module " << Asc_ModuleName(m) << ": " << msg1 << std::endl;
	}

	MSG("Beginning parse of %s",Asc_ModuleName(m));
#ifdef LOADSTRING_ERROR_TREE
	error_reporter_tree_t *tree1 = error_reporter_tree_start(0);
#endif
	status = zz_parse();
	switch(status){
		case 0: break;
		case 1: ERROR_REPORTER_NOLINE(ASC_USER_ERROR,"Parsing of %s was aborted",Asc_ModuleName(m)); break;
		case 2: ERROR_REPORTER_NOLINE(ASC_PROG_FATAL,"Out of memory when parsing %s",Asc_ModuleName(m)); break;
		default: ERROR_REPORTER_NOLINE(ASC_PROG_ERROR,"Invalid return from zz_parse"); break;
	}
#ifdef LOADSTRING_ERROR_TREE
	status = error_reporter_tree_has_error(tree1);
	if(!status){
		MSG("CLEARING TREE...");
		error_reporter_tree_end_clear(tree1);
		MSG("DONE CLEARING TREE...");
	}else{
		error_reporter_tree_end(tree1);
		ERROR_REPORTER_NOLINE(ASC_USER_ERROR,"Error(s) when loading '%s'",nameprefix);
		stringstream ss;
		ss << "Errors found in '" << nameprefix <<  "'";
		throw runtime_error(ss.str());
	}
#endif

	struct gl_list_t *l = Asc_TypeByModule(m);
	MSG("%lu library entries loaded from %s",gl_length(l), nameprefix);
}

const char *
Library::getLoadErrorMessage(int status){
	const char *msg = NULL;
	switch(status){
		case 5:
			msg = "The module '%s' already exists. "; break;
		case 4:
			msg = "Caught an attempt to do a recursive require under '%s'."; break;
		case 3:
			msg = "A new module was created from '%s', overwriting a module's alias."; break;
		case 2:
			msg = "An existing module is being returned for '%s'." ; break;
		case 1:
			msg = "An new version of an existing module was created for '%s'."; break;
		case 0:
			msg = "Module for '%s' created OK."; break;
		case -1:
			msg = "File not found for '%s'. (-1)"; break;
		case -2:
			msg = "Unable to open '%s' for reading. (-2)";break;
		case -3:
			msg = "Insuffient memory to create module for '%s'. (-3)"; break;
		case -4:
			msg = "Bad input, null or zero length filename in '%s'. (-4)"; break;
		default:
			throw std::runtime_error("Invalid status code in library.cpp");
	}
	return msg;
}





/**
	Return a vector of all the Modules which have been loaded into
	the current library.
*/
vector<Module>
Library::getModules(const int module_type){
	if(module_type < 0 || module_type > 2){
		throw std::runtime_error("Library::getModules: invalid module_type parameter");
	}

	vector<Module> v;
	struct gl_list_t *l = Asc_ModuleList(module_type);
	if(l==NULL){
		MSG("list is empty");
		return v;
	}
	for(int i=0, end=gl_length(l); i<end; ++i){
		symchar *name = (symchar *)gl_fetch(l,i+1);
		if(AscFindSymbol(name)==NULL){
			throw runtime_error("Library::getModules: invalid symchar *");
		}
		//cerr << "GOT " << SCP( name ) << endl;
		const module_t *m = Asc_GetModuleByName((const char *)name);
		v.push_back(Module(m));
	}
	/*cerr << "LENGTH OF V IS " << v.size() << endl;
	if(v.size()){
		cerr << "MODULE 0's NAME IS " << v[0].getName() << endl;
	}*/
	return v;
}

/**
	Output to stderr the names of the modules loaded into the current Library.
*/
void
Library::listModules(const int module_type){
	if(module_type < 0 || module_type > 2){
		throw std::runtime_error("Library::listModules: invalid module_type parameter");
	}

	const char *type = NULL;
	switch(module_type){
		case 0: type = "defined types"; break;
		case 1: type = "string definitions"; break;
		case 2: type = "statements"; break;
	}

	vector<Module> v = getModules(module_type);
	if(v.size()){
		std::cerr << "Listing " << v.size() << " modules with " << type << std::endl;
		for(vector<Module>::const_iterator i=v.begin(); i < v.end(); ++i){
			cerr << " - " << i->getName() << endl;
			vector<Type> tt = getModuleTypes(*i);
			for(vector<Type>::const_iterator j = tt.begin(); j < tt.end(); ++j){
				cerr << "    * " << j->getName() << endl;
			}
		}
	}else{
		std::cerr << "Notice: No " << type << " found in module list." << std::endl;
	}
}

Type &
Library::findType(const SymChar &sym){
	TypeDescription *t = FindType(sym.getInternalType());
	if(t==NULL){
		stringstream ss;
		ss << "Library::findType: type '" << sym << "' not found in library";
		throw runtime_error(ss.str());
	}/*else{
		cerr << "Found something for type " << sym << endl;
	}*/
	Type *t2=new Type(t);
	return *t2;
}

static int module_sort_cmp(const void *a, const void *b){
	return strcmp((char *)a, (char *)b);
}

/**
	This could be quite a bit more efficient if we could get a gl_list_t of TypeDescription rather than names
*/
vector<Type>
Library::getModuleTypes(const Module &m){
	//cerr << "GET MODULE TYPES\n" << endl;
	vector<Type> v;
	struct gl_list_t *l =  Asc_TypeByModule(m.getInternalType());

	gl_sort(l, &module_sort_cmp);

	for(int i=0,end=gl_length(l); i<end; ++i){
		char *name = (char *)gl_fetch(l,i+1);
		MSG("Found type %s",name);
		TypeDescription *t = FindType((const symchar *)name);
		v.push_back(Type(t));
	}
	return v;
}

/**
	This function is kinda fighting against the Table implementation of the external function library. What we really need is some kind of iterator on the Table struct, but it doesn't seem to be implemented. Instead there is a C-style equivalent of the STL 'bind1st' function which we can use, but it's not exported from the current extfunc.h so we need to add it.
*/
vector<ExtMethod>
Library::getExtMethods(){
	// Clear the vector
	extmethod_vector = vector<ExtMethod>();

	// Traverse the vector
	TraverseExtFuncLibrary(Library::extMethodTraverse, (void *)this);

	return extmethod_vector;
}

/**
	This method exists only to allow the TraverseExtFuncLibrary function
	to make callbacks to the Library class from C.

	@NOTE there might be issues with C/C++ linking here?
*/
void
Library::extMethodTraverse(void *a1, void *a2){
	Library *self = (Library *)a2;
	self->appendToExtMethodVector(a1);
}

void
Library::appendToExtMethodVector(void *a1){
	struct ExternalFunc *e = (struct ExternalFunc *)a1;
	extmethod_vector.push_back(ExtMethod(e));
}

/**
	Clear the library: 'DESTROY TYPES'

	@TODO do this more efficiently, don't destroy the whole ASCEND compiler.
*/
void
Library::clear(){
/*	Asc_CompilerDestroy();
	cerr << "COMPLETED ASC_COMPILERDESTROY" << endl;
	Asc_CompilerInit(1);
	cerr << "... ASC_COMPILERINIT OK" << endl;
	Asc_ImportPathList(PATHENVIRONMENTVAR);
	registerStandardSolvers();
	cerr << "... REGISTER_STANDARD_SOLVERS" << endl;
	DefineFundamentalTypes();
	cerr << "... DEFINED FUND TYPES" << endl;
	\*SetUniversalProcedureList(NULL);
*/

	MSG("Displaying library modules and types...");
	//listModules();

	MSG("Destroying simulations...");
	Asc_DestroySimulations();

	MSG("Clearing library...");
	DestroyNotesDatabase(LibraryNote());
	SetUniversalProcedureList(NULL);
	DestroyLibrary();
	DestroyPrototype();
	EmptyTrash();
	Asc_DestroyModules((DestroyFunc)DestroyStatementList);
	//importhandler_destroylibrary();
	WriteChildMissing(NULL,NULL,NULL);
	//Asc_CompilerInit(1)
	DefineFundamentalTypes();
	InitNotesDatabase(LibraryNote());
	ERROR_REPORTER_NOLINE(ASC_PROG_WARNING,"LIBRARY CLEARED!");
}

AnnotationDatabase
Library::getAnnotationDatabase(){
	return AnnotationDatabase(SCP(LibraryNote()));
}

vector<UnitsM>
Library::getUnits() const{
	vector<UnitsM> v;
    register unsigned long c;
    const struct Units *p;
    for(c = 0;c<UNITS_HASH_SIZE;c++) {
		for(p = g_units_hash_table[c];p!=NULL;p = p->next){
			v.push_back(UnitsM(p));
		}
	}
	return v;
}

set<Type>
Library::getRealAtomTypes() const{
	set<Type> s;
	struct gl_list_t *l = DefinitionList();
	if(!l){
		throw runtime_error("No types found in library (perhaps no files have been loaded?)");
	}

	for(unsigned i = 1; i<=gl_length(l); ++i) {
		Type t((const struct TypeDescription *)gl_fetch(l,i));
		if(t.isRefinedReal()){
			Dimensions d = t.getDimensions();
			if(d.isWild() || d.isDimensionless())continue; // skip this one
			// it's got some dimensions, add it to the list			
			s.insert(t);
		}
	}
	gl_destroy(l);
	return s;
}

